@@ -59,8 +59,8 @@ static void *hello_init(struct fuse_conn_info *conn,
 	cfg->kernel_cache = 1;
 
 	/* Test setting flags the old way */
-	fuse_set_feature_flag(conn, FUSE_CAP_ASYNC_READ);
-	fuse_unset_feature_flag(conn, FUSE_CAP_ASYNC_READ);
+	conn->want = FUSE_CAP_ASYNC_READ;
+	conn->want &= ~FUSE_CAP_ASYNC_READ;
 
 	return NULL;
 }
