@@ -13,7 +13,6 @@
 #include <stdint.h>
 #include <stdbool.h>
 #include <errno.h>
-#include <stdatomic.h>
 
 #define MIN(a, b) \
 ({									\
@@ -55,7 +54,7 @@ struct fuse_notify_req {
 };
 
 struct fuse_session {
-	_Atomic(char *)mountpoint;
+	char *mountpoint;
 	volatile int exited;
 	int fd;
 	struct fuse_custom_io *io;
@@ -76,23 +75,14 @@ struct fuse_session {
 	int broken_splice_nonblock;
 	uint64_t notify_ctr;
 	struct fuse_notify_req notify_list;
-	_Atomic size_t bufsize;
+	size_t bufsize;
 	int error;
 
 	/* This is useful if any kind of ABI incompatibility is found at
 	 * a later version, to 'fix' it at run time.
 	 */
 	struct libfuse_version version;
-
-	/* true if reading requests from /dev/fuse are handled internally */
 	bool buf_reallocable;
-
-	/*
-	 * conn->want and conn_want_ext options set by libfuse , needed
-	 * to correctly convert want to want_ext
-	 */
-	uint32_t conn_want;
-	uint64_t conn_want_ext;
 };
 
 struct fuse_chan {
@@ -235,3 +225,34 @@ int fuse_loop_cfg_verify(struct fuse_loop_config *config);
 /* room needed in buffer to accommodate header */
 #define FUSE_BUFFER_HEADER_SIZE 0x1000
 
+/**
+ * Get the wanted capability flags, converting from old format if necessary
+ */
+static inline int convert_to_conn_want_ext(struct fuse_conn_info *conn,
+					   uint64_t want_ext_default,
+					   uint32_t want_default)
+{
+	/*
+	 * Convert want to want_ext if necessary.
+	 * For the high level interface this function might be called
+	 * twice, once from the high level interface and once from the
+	 * low level interface. Both, with different want_ext_default and
+	 * want_default values. In order to suppress a failure for the
+	 * second call, we check if the lower 32 bits of want_ext are
+	 * already set to the value of want.
+	 */
+	if (conn->want != want_default &&
+	    fuse_lower_32_bits(conn->want_ext) != conn->want) {
+		if (conn->want_ext != want_ext_default) {
+			fuse_log(FUSE_LOG_ERR,
+				 "fuse: both 'want' and 'want_ext' are set\n");
+			return -EINVAL;
+		}
+
+		/* high bits from want_ext, low bits from want */
+		conn->want_ext = fuse_higher_32_bits(conn->want_ext) |
+				 conn->want;
+	}
+
+	return 0;
+}
