@@ -5,7 +5,7 @@
   Implementation of the multi-threaded FUSE session loop.
 
   This program can be distributed under the terms of the GNU LGPLv2.
-  See the file LGPL2.txt.
+  See the file COPYING.LIB.
 */
 
 #define _GNU_SOURCE
@@ -15,7 +15,6 @@
 #include "fuse_misc.h"
 #include "fuse_kernel.h"
 #include "fuse_i.h"
-#include "fuse_uring_i.h"
 #include "util.h"
 
 #include <stdio.h>
@@ -54,12 +53,14 @@ struct fuse_worker {
 	struct fuse_mt *mt;
 };
 
-/* synchronization via se->mt_lock */
 struct fuse_mt {
+	pthread_mutex_t lock;
 	int numworker;
 	int numavail;
 	struct fuse_session *se;
 	struct fuse_worker main;
+	sem_t finish;
+	int exit;
 	int error;
 	int clone_fd;
 	int max_idle;
@@ -130,30 +131,30 @@ static void *fuse_do_work(void *data)
 {
 	struct fuse_worker *w = (struct fuse_worker *) data;
 	struct fuse_mt *mt = w->mt;
-	struct fuse_session *se = mt->se;
 
-	fuse_set_thread_name("fuse_worker");
+	pthread_setname_np(pthread_self(), "fuse_worker");
 
-	while (!fuse_session_exited(se)) {
+	while (!fuse_session_exited(mt->se)) {
 		int isforget = 0;
 		int res;
 
 		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
-		res = fuse_session_receive_buf_internal(se, &w->fbuf, w->ch);
+		res = fuse_session_receive_buf_internal(mt->se, &w->fbuf,
+							w->ch);
 		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
 		if (res == -EINTR)
 			continue;
 		if (res <= 0) {
 			if (res < 0) {
-				fuse_session_exit(se);
+				fuse_session_exit(mt->se);
 				mt->error = res;
 			}
 			break;
 		}
 
-		pthread_mutex_lock(&se->mt_lock);
-		if (fuse_session_exited(se)) {
-			pthread_mutex_unlock(&se->mt_lock);
+		pthread_mutex_lock(&mt->lock);
+		if (mt->exit) {
+			pthread_mutex_unlock(&mt->lock);
 			return NULL;
 		}
 
@@ -171,14 +172,13 @@ static void *fuse_do_work(void *data)
 
 		if (!isforget)
 			mt->numavail--;
-		if (mt->numavail == 0 && mt->numworker < mt->max_threads &&
-		    likely(se->got_init))
+		if (mt->numavail == 0 && mt->numworker < mt->max_threads)
 			fuse_loop_start_thread(mt);
-		pthread_mutex_unlock(&se->mt_lock);
+		pthread_mutex_unlock(&mt->lock);
 
-		fuse_session_process_buf_internal(se, &w->fbuf, w->ch);
+		fuse_session_process_buf_internal(mt->se, &w->fbuf, w->ch);
 
-		pthread_mutex_lock(&se->mt_lock);
+		pthread_mutex_lock(&mt->lock);
 		if (!isforget)
 			mt->numavail++;
 
@@ -189,14 +189,14 @@ static void *fuse_do_work(void *data)
 		 * delayed, a moving average might be useful for that.
 		 */
 		if (mt->max_idle != -1 && mt->numavail > mt->max_idle && mt->numworker > 1) {
-			if (fuse_session_exited(se)) {
-				pthread_mutex_unlock(&se->mt_lock);
+			if (mt->exit) {
+				pthread_mutex_unlock(&mt->lock);
 				return NULL;
 			}
 			list_del_worker(w);
 			mt->numavail--;
 			mt->numworker--;
-			pthread_mutex_unlock(&se->mt_lock);
+			pthread_mutex_unlock(&mt->lock);
 
 			pthread_detach(w->thread_id);
 			fuse_buf_free(&w->fbuf);
@@ -204,10 +204,11 @@ static void *fuse_do_work(void *data)
 			free(w);
 			return NULL;
 		}
-		pthread_mutex_unlock(&se->mt_lock);
+		pthread_mutex_unlock(&mt->lock);
 	}
 
-	sem_post(&se->mt_finish);
+	sem_post(&mt->finish);
+
 	return NULL;
 }
 
@@ -272,15 +273,9 @@ static int fuse_clone_chan_fd_default(struct fuse_session *se)
 			strerror(errno));
 		return -1;
 	}
-	if (!O_CLOEXEC) {
-		res = fcntl(clonefd, F_SETFD, FD_CLOEXEC);
-		if (res == -1) {
-			fuse_log(FUSE_LOG_ERR, "fuse: failed to set CLOEXEC: %s\n",
-				strerror(errno));
-			close(clonefd);
-			return -1;
-		}
-	}
+#ifndef O_CLOEXEC
+	fcntl(clonefd, F_SETFD, FD_CLOEXEC);
+#endif
 
 	masterfd = se->fd;
 	res = ioctl(clonefd, FUSE_DEV_IOC_CLONE, &masterfd);
@@ -357,9 +352,9 @@ static int fuse_loop_start_thread(struct fuse_mt *mt)
 static void fuse_join_worker(struct fuse_mt *mt, struct fuse_worker *w)
 {
 	pthread_join(w->thread_id, NULL);
-	pthread_mutex_lock(&mt->se->mt_lock);
+	pthread_mutex_lock(&mt->lock);
 	list_del_worker(w);
-	pthread_mutex_unlock(&mt->se->mt_lock);
+	pthread_mutex_unlock(&mt->lock);
 	fuse_buf_free(&w->fbuf);
 	fuse_chan_put(w->ch);
 	free(w);
@@ -395,35 +390,34 @@ int err;
 	mt.max_threads = config->max_threads;
 	mt.main.thread_id = pthread_self();
 	mt.main.prev = mt.main.next = &mt.main;
+	sem_init(&mt.finish, 0, 0);
+	pthread_mutex_init(&mt.lock, NULL);
 
-	pthread_mutex_lock(&se->mt_lock);
+	pthread_mutex_lock(&mt.lock);
 	err = fuse_loop_start_thread(&mt);
-	pthread_mutex_unlock(&se->mt_lock);
+	pthread_mutex_unlock(&mt.lock);
 	if (!err) {
+		/* sem_wait() is interruptible */
 		while (!fuse_session_exited(se))
-			sem_wait(&se->mt_finish);
-		if (se->debug)
-			fuse_log(FUSE_LOG_DEBUG,
-				 "fuse: session exited, terminating workers\n");
+			sem_wait(&mt.finish);
 
-		pthread_mutex_lock(&se->mt_lock);
+		pthread_mutex_lock(&mt.lock);
 		for (w = mt.main.next; w != &mt.main; w = w->next)
 			pthread_cancel(w->thread_id);
-		pthread_mutex_unlock(&se->mt_lock);
+		mt.exit = 1;
+		pthread_mutex_unlock(&mt.lock);
 
 		while (mt.main.next != &mt.main)
 			fuse_join_worker(&mt, mt.main.next);
 
 		err = mt.error;
-
-		if (se->uring.pool)
-			fuse_uring_stop(se);
 	}
 
-	pthread_mutex_destroy(&se->mt_lock);
+	pthread_mutex_destroy(&mt.lock);
+	sem_destroy(&mt.finish);
 	if(se->error != 0)
 		err = se->error;
-
+	fuse_session_reset(se);
 
 	if (created_config) {
 		fuse_loop_cfg_destroy(config);
