@@ -44,6 +44,10 @@
 #define PARAM(inarg) (((char *)(inarg)) + sizeof(*(inarg)))
 #define OFFSET_MAX 0x7fffffffffffffffLL
 
+#define container_of(ptr, type, member) ({				\
+			const typeof( ((type *)0)->member ) *__mptr = (ptr); \
+			(type *)( (char *)__mptr - offsetof(type,member) );})
+
 struct fuse_pollhandle {
 	uint64_t kh;
 	struct fuse_session *se;
@@ -1994,77 +1998,6 @@ static bool want_flags_valid(uint64_t capable, uint64_t want)
 	return true;
 }
 
-/**
- * Get the wanted capability flags, converting from old format if necessary
- */
-int fuse_convert_to_conn_want_ext(struct fuse_conn_info *conn)
-{
-	struct fuse_session *se = container_of(conn, struct fuse_session, conn);
-
-	/*
-	 * Convert want to want_ext if necessary.
-	 * For the high level interface this function might be called
-	 * twice, once from the high level interface and once from the
-	 * low level interface. Both, with different want_ext_default and
-	 * want_default values. In order to suppress a failure for the
-	 * second call, we check if the lower 32 bits of want_ext are
-	 * already set to the value of want.
-	 */
-	if (conn->want != se->conn_want &&
-	    fuse_lower_32_bits(conn->want_ext) != conn->want) {
-		if (conn->want_ext != se->conn_want_ext) {
-			fuse_log(FUSE_LOG_ERR,
-				"%s: Both conn->want_ext and conn->want are set.\n"
-				"want=%x, want_ext=%lx, se->want=%lx se->want_ext=%lx\n",
-				__func__, conn->want, conn->want_ext,
-				se->conn_want, se->conn_want_ext);
-			return -EINVAL;
-		}
-
-		/* high bits from want_ext, low bits from want */
-		conn->want_ext = fuse_higher_32_bits(conn->want_ext) |
-				 conn->want;
-	}
-
-	/* ensure there won't be a second conversion */
-	conn->want = fuse_lower_32_bits(conn->want_ext);
-
-	return 0;
-}
-
-bool fuse_set_feature_flag(struct fuse_conn_info *conn,
-					 uint64_t flag)
-{
-	struct fuse_session *se = container_of(conn, struct fuse_session, conn);
-
-	if (conn->capable_ext & flag) {
-		conn->want_ext |= flag;
-		se->conn_want_ext |= flag;
-		conn->want  |= flag;
-		se->conn_want |= flag;
-		return true;
-	}
-	return false;
-}
-
-void fuse_unset_feature_flag(struct fuse_conn_info *conn,
-					 uint64_t flag)
-{
-	struct fuse_session *se = container_of(conn, struct fuse_session, conn);
-
-	conn->want_ext &= ~flag;
-	se->conn_want_ext &= ~flag;
-	conn->want  &= ~flag;
-	se->conn_want &= ~flag;
-}
-
-bool fuse_get_feature_flag(struct fuse_conn_info *conn,
-					     uint64_t flag)
-{
-	return conn->capable_ext & flag ? true : false;
-}
-
-
 /* Prevent bogus data races (bogus since "init" is called before
  * multi-threading becomes relevant */
 static __attribute__((no_sanitize("thread")))
@@ -2225,8 +2158,13 @@ void do_init(fuse_req_t req, fuse_ino_t nodeid, const void *inarg)
 
 	se->got_init = 1;
 	if (se->op.init) {
+		uint64_t want_ext_default = se->conn.want_ext;
+		uint32_t want_default = fuse_lower_32_bits(se->conn.want_ext);
+		int rc;
+
 		// Apply the first 32 bits of capable_ext to capable
 		se->conn.capable = fuse_lower_32_bits(se->conn.capable_ext);
+		se->conn.want = want_default;
 
 		se->op.init(se->userdata, &se->conn);
 
@@ -2235,7 +2173,14 @@ void do_init(fuse_req_t req, fuse_ino_t nodeid, const void *inarg)
 		 * se->conn.want_ext
 		 * Userspace might still use conn.want - we need to convert it
 		 */
-		fuse_convert_to_conn_want_ext(&se->conn);
+		rc = convert_to_conn_want_ext(&se->conn, want_ext_default,
+					      want_default);
+		if (rc != 0) {
+			fuse_reply_err(req, EPROTO);
+			se->error = -EPROTO;
+			fuse_session_exit(se);
+			return;
+		}
 	}
 
 	if (!want_flags_valid(se->conn.capable_ext, se->conn.want_ext)) {
@@ -2378,14 +2323,10 @@ void do_init(fuse_req_t req, fuse_ino_t nodeid, const void *inarg)
 static void do_destroy(fuse_req_t req, fuse_ino_t nodeid, const void *inarg)
 {
 	struct fuse_session *se = req->se;
-	char *mountpoint;
 
 	(void) nodeid;
 	(void) inarg;
 
-	mountpoint = atomic_exchange(&se->mountpoint, NULL);
-	free(mountpoint);
-
 	se->got_destroy = 1;
 	se->got_init = 0;
 	if (se->op.destroy)
@@ -3085,14 +3026,11 @@ static int _fuse_session_receive_buf(struct fuse_session *se,
 {
 	int err;
 	ssize_t res;
-	size_t bufsize;
+	size_t bufsize = se->bufsize;
 #ifdef HAVE_SPLICE
 	struct fuse_ll_pipe *llp;
 	struct fuse_buf tmpbuf;
 
-pipe_retry:
-	bufsize = se->bufsize;
-
 	if (se->conn.proto_minor < 14 ||
 	    !(se->conn.want_ext & FUSE_CAP_SPLICE_READ))
 		goto fallback;
@@ -3137,13 +3075,6 @@ pipe_retry:
 			fuse_session_exit(se);
 			return 0;
 		}
-
-		/* FUSE_INIT might have increased the required bufsize */
-		if (err == EINVAL && bufsize < se->bufsize) {
-			fuse_ll_clear_pipe(se);
-			goto pipe_retry;
-		}
-
 		if (err != EINTR && err != EAGAIN)
 			perror("fuse: splice from device");
 		return -err;
@@ -3171,16 +3102,18 @@ pipe_retry:
 		struct fuse_bufvec dst = { .count = 1 };
 
 		if (!buf->mem) {
-			buf->mem = buf_alloc(bufsize, internal);
+			buf->mem = buf_alloc(se->bufsize, internal);
 			if (!buf->mem) {
 				fuse_log(
 					FUSE_LOG_ERR,
 					"fuse: failed to allocate read buffer\n");
 				return -ENOMEM;
 			}
-			buf->mem_size = bufsize;
+			buf->mem_size = se->bufsize;
+			if (internal)
+				se->buf_reallocable = true;
 		}
-		buf->size = bufsize;
+		buf->size = se->bufsize;
 		buf->flags = 0;
 		dst.buf[0] = *buf;
 
@@ -3210,21 +3143,21 @@ pipe_retry:
 
 fallback:
 #endif
-	bufsize = internal ? buf->mem_size : se->bufsize;
 	if (!buf->mem) {
-		bufsize = se->bufsize; /* might have changed */
-		buf->mem = buf_alloc(bufsize, internal);
+		buf->mem = buf_alloc(se->bufsize, internal);
 		if (!buf->mem) {
 			fuse_log(FUSE_LOG_ERR,
 				 "fuse: failed to allocate read buffer\n");
 			return -ENOMEM;
 		}
-
+		buf->mem_size = se->bufsize;
 		if (internal)
-			buf->mem_size = bufsize;
+			se->buf_reallocable = true;
 	}
 
 restart:
+	if (se->buf_reallocable)
+		bufsize = buf->mem_size;
 	if (se->io != NULL) {
 		/* se->io->read is never NULL if se->io is not NULL as
 		specified by fuse_session_custom_io()*/
@@ -3238,10 +3171,9 @@ restart:
 	if (fuse_session_exited(se))
 		return 0;
 	if (res == -1) {
-		if (err == EINVAL && internal && se->bufsize > bufsize) {
-			/* FUSE_INIT might have increased the required bufsize */
-			bufsize = se->bufsize;
-			void *newbuf = buf_alloc(bufsize, internal);
+		if (err == EINVAL && se->buf_reallocable &&
+		    se->bufsize > buf->mem_size) {
+			void *newbuf = buf_alloc(se->bufsize, internal);
 			if (!newbuf) {
 				fuse_log(
 					FUSE_LOG_ERR,
@@ -3250,7 +3182,8 @@ restart:
 			}
 			fuse_buf_free(buf);
 			buf->mem = newbuf;
-			buf->mem_size = bufsize;
+			buf->mem_size = se->bufsize;
+			se->buf_reallocable = true;
 			goto restart;
 		}
 
@@ -3292,13 +3225,6 @@ int fuse_session_receive_buf_internal(struct fuse_session *se,
 				      struct fuse_buf *buf,
 				      struct fuse_chan *ch)
 {
-	/*
-	 * if run internally thread buffers are from libfuse - we can
-	 * reallocate them
-	 */
-	if (unlikely(!se->got_init) && !se->buf_reallocable)
-		se->buf_reallocable = true;
-
 	return _fuse_session_receive_buf(se, buf, ch, true);
 }
 
@@ -3470,20 +3396,12 @@ int fuse_session_custom_io_30(struct fuse_session *se,
 			offsetof(struct fuse_custom_io, clone_fd), fd);
 }
 
-int fuse_session_mount(struct fuse_session *se, const char *_mountpoint)
+int fuse_session_mount(struct fuse_session *se, const char *mountpoint)
 {
 	int fd;
-	char *mountpoint;
 
-	if (_mountpoint == NULL) {
-		fuse_log(FUSE_LOG_ERR, "Invalid null-ptr mountpoint!\n");
-		return -1;
-	}
-
-	mountpoint = strdup(_mountpoint);
 	if (mountpoint == NULL) {
-		fuse_log(FUSE_LOG_ERR, "Failed to allocate memory for mountpoint. Error: %s\n",
-			strerror(errno));
+		fuse_log(FUSE_LOG_ERR, "Invalid null-ptr mountpoint!\n");
 		return -1;
 	}
 
@@ -3509,7 +3427,7 @@ int fuse_session_mount(struct fuse_session *se, const char *_mountpoint)
 			fuse_log(FUSE_LOG_ERR,
 				"fuse: Invalid file descriptor /dev/fd/%u\n",
 				fd);
-			goto error_out;
+			return -1;
 		}
 		se->fd = fd;
 		return 0;
@@ -3518,16 +3436,18 @@ int fuse_session_mount(struct fuse_session *se, const char *_mountpoint)
 	/* Open channel */
 	fd = fuse_kern_mount(mountpoint, se->mo);
 	if (fd == -1)
-		goto error_out;
+		return -1;
 	se->fd = fd;
 
 	/* Save mountpoint */
-	se->mountpoint = mountpoint;
+	se->mountpoint = strdup(mountpoint);
+	if (se->mountpoint == NULL)
+		goto error_out;
 
 	return 0;
 
 error_out:
-	free(mountpoint);
+	fuse_kern_unmount(mountpoint, fd);
 	return -1;
 }
 
@@ -3539,11 +3459,10 @@ int fuse_session_fd(struct fuse_session *se)
 void fuse_session_unmount(struct fuse_session *se)
 {
 	if (se->mountpoint != NULL) {
-		char *mountpoint = atomic_exchange(&se->mountpoint, NULL);
-
-		fuse_kern_unmount(mountpoint, se->fd);
+		fuse_kern_unmount(se->mountpoint, se->fd);
 		se->fd = -1;
-		free(mountpoint);
+		free(se->mountpoint);
+		se->mountpoint = NULL;
 	}
 }
 
