@@ -5,7 +5,7 @@
   Utility functions for setting signal handlers.
 
   This program can be distributed under the terms of the GNU LGPLv2.
-  See the file LGPL2.txt
+  See the file COPYING.LIB
 */
 
 #include "fuse_config.h"
@@ -22,20 +22,13 @@
 #include <execinfo.h>
 #endif
 
-/*
- * Must not handle SIGCANCEL, as that is used to wake up threads from
- * syscalls reading requests from /dev/fuse
- */
 static int teardown_sigs[] = { SIGHUP, SIGINT, SIGTERM };
-
 static int ignore_sigs[] = { SIGPIPE};
 static int fail_sigs[] = { SIGILL, SIGTRAP, SIGABRT, SIGBUS, SIGFPE, SIGSEGV };
 static struct fuse_session *fuse_instance;
 
-#ifdef HAVE_BACKTRACE
 #define BT_STACK_SZ (1024 * 1024)
 static void *backtrace_buffer[BT_STACK_SZ];
-#endif
 
 static void dump_stack(void)
 {
@@ -60,14 +53,8 @@ static void dump_stack(void)
 
 static void exit_handler(int sig)
 {
-	if (fuse_instance == NULL) {
-		fuse_log(FUSE_LOG_ERR, "fuse_instance is NULL\n");
+	if (fuse_instance == NULL)
 		return;
-	}
-
-	if (fuse_instance->debug)
-		fuse_log(FUSE_LOG_ERR, "exit_handler called with sig %d\n",
-			 sig);
 
 	fuse_session_exit(fuse_instance);
 
@@ -163,13 +150,8 @@ int fuse_set_signal_handlers(struct fuse_session *se)
 	if (rc < 0)
 		return rc;
 
-	/*
-	 * needs to be set independently if already set, as some  applications
-	 * may have multiple sessions and might rely on traditional behavior
-	 * that the last session is used.
-	 */
-	fuse_instance = se;
-
+	if (fuse_instance == NULL)
+		fuse_instance = se;
 	return 0;
 }
 
@@ -182,8 +164,8 @@ int fuse_set_fail_signal_handlers(struct fuse_session *se)
 	if (rc < 0)
 		return rc;
 
-	/* See fuse_set_signal_handlers, why set unconditionally */
-	fuse_instance = se;
+	if (fuse_instance == NULL)
+		fuse_instance = se;
 
 	return 0;
 }
